{+--------------------------------------------------------------------------+
 | Class:       TmTSearchO  ( TurboSearch )
 | Created:     8.97
 | Author:      Martin Waldenburg
 | Description: A very fast search engine, about twice as fast as Boyer-Moore,
 |              based on an article in the German magazine c't (8/97).
 |              The original is in 'C '. www.Heise.com.
 |              You can search case sensitive or case insensitive.
 |              Look_At is implemented.
 | Version:     2.2
 | Copyright (c) 1997 - 1998 Martin Waldenburg
 | All rights reserved.
 |
 | LICENCE CONDITIONS
 |
 | USE OF THE ENCLOSED SOFTWARE
 | INDICATES YOUR ASSENT TO THE
 | FOLLOWING LICENCE CONDITIONS.
 |
 |
 |
 | These Licence Conditions are exlusively
 | governed by the Law and Rules of the
 | Federal Republic of Germany.
 |
 | Redistribution and use in source and binary form, with or without
 | modification, are permitted provided that the following conditions
 | are met:
 |
 | 1. Software using this code must contain a visible line of credit,
 |    to Martin Waldenburg and the German magazine c't (8/97).
 |
 | DISCLAIMER:
 |
 | THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS IS'.
 |
 | ALL EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 | PARTICULAR PURPOSE ARE DISCLAIMED.
 |
 | IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 | INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 | (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 | OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 | INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 | WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 | NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 | THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 |
 |  Martin.Waldenburg@T-Online.de

+--------------------------------------------------------------------------+}
unit mwTSearchO;


interface

uses
  Windows,
  SysUtils,
  Messages,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  Menus;

type
  TmTSearchO = class(Tobject)
  private
    Text, Pat: String;
    fCount: Integer;
    Look_At: Integer;
    fPos: Integer;
    PatLen, PatLenPlus, TextLen: Integer;
    Shift: array[0..255] of Integer;
    fFinished: Boolean;
    fFound: Boolean;
    function GetFinished: Boolean;
  protected
  public
    constructor Create;
    destructor Destroy; override;
    procedure Init(NewPattern: String);
    function FindFirst(NewText: String): Integer;
    function Next: Integer;
    procedure IInit(NewPattern: String);
    function IFindFirst(NewText: String): Integer;
    function INext: Integer;
    property Count: Integer read fCount write fCount;
    property Finished: Boolean read GetFinished;
    property Found: Boolean read fFound;
  published
  end;

var
  CompTable: array[#0..#255] of byte;

implementation

constructor TmTSearchO.Create;
begin
  inherited Create;
  Pat := '';
  PatLen := 0;
  Text := '';
  TextLen := 0;
  fPos := 0;
  fFound := False;
  fCount := 0;
end; { Create }

procedure MakeCompTable;
var
  I: Byte;
begin
  for I := 0 to 255 do
    CompTable[Char(I)] := ord(AnsiLowerCase(Char(I))[1]);
end; { MakeCompTable }

function TmTSearchO.GetFinished: Boolean;
begin
  fFinished := False;
  if fPos >= TextLen then fFinished := True;
  if PatLen > TextLen then fFinished := True;
  Result := fFinished;
end; { GetFinished }

procedure TmTSearchO.Init(NewPattern: String);
var
  I: Byte;
begin
  Pat := NewPattern;
  PatLen := Length(Pat);
  PatLenPlus := PatLen + 1;
  Look_At := 1;
  for I := 0 to 255 do Shift[I] := PatLenPlus;
  for I := 1 to PatLen do Shift[ord(Pat[I])] := PatLenPlus - I;
  while Look_at < PatLen do
  begin
    if Pat[PatLen] = Pat[PatLen - (Look_at)] then exit;
    inc(Look_at);
  end;
end; { Init }

function TmTSearchO.FindFirst(NewText: String): Integer;
begin
  Text := NewText;
  TextLen := Length(Text);
  Result := 0;
  fFound := False;
  fCount := 0;
  if TextLen >= PatLen then
  begin
    fPos := 0;
    Result := Next;
  end;
end; { FindFirst }

function TmTSearchO.Next: Integer;
var
  I, J: Integer;
begin
  Result := 0;
  fFound := False;
  inc(fPos, PatLen);
  while fPos <= TextLen do
  begin
    if (Pat[Patlen] <> Text[fPos]) then inc(fPos, Shift[ord(Text[fPos + 1])])
    else
    begin
      J := fPos - PatLen + 1;
      I := 1;
      while Pat[I] = Text[J] do
      begin
        inc(I);
        if I = Patlen then
        begin
          fFound := True;
          inc(fCount);
          Result := fPos - Patlen + 1;
          exit;
        end;
        inc(J);
      end;
      inc(fPos, Look_at + Shift[ord(Text[fPos + Look_at])] - 1); // Suggestion of Angus Johnson
    end;
  end;
end; { Next }

procedure TmTSearchO.IInit(NewPattern: String);
var
  I: Byte;
begin
  Pat := NewPattern;
  PatLen := Length(Pat);
  PatLenPlus := PatLen + 1;
  Look_At := 1;
  for I := 0 to 255 do Shift[I] := PatLenPlus;
  for I := 1 to PatLen do Shift[CompTable[Pat[i]]] := PatLenPlus - I;
  while Look_at < PatLen do
  begin
    if CompTable[Pat[PatLen]] = CompTable[Pat[PatLen - (Look_at)]] then exit;
    inc(Look_at);
  end;
end; { IInit }

function TmTSearchO.IFindFirst(NewText: String): Integer;
begin
  Text := NewText;
  TextLen := Length(Text);
  Result := 0;
  fFound := False;
  fCount := 0;
  if TextLen >= PatLen then
  begin
    fPos := 0;
    Result := INext;
  end;
end; { IFindFirst }

function TmTSearchO.INext: Integer;
var
  I, J: Integer;
begin
  Result := 0;
  fFound := False;
  inc(fPos, PatLen);
  while fPos <= TextLen do
  begin
    if (CompTable[Pat[Patlen]] <> CompTable[Text[fPos]]) then
      inc(fPos, Shift[CompTable[Text[fPos + 1]]])
    else
    begin
      J := fPos - PatLen + 1;
      I := 1;
      while CompTable[Pat[I]] = CompTable[Text[J]] do
      begin
        inc(I);
        if I = Patlen then
        begin
          fFound := True;
          inc(fCount);
          Result := fPos - Patlen + 1;
          exit;
        end;
        inc(J);
      end;
      inc(fPos, Look_At + Shift[CompTable[Text[fPos + Look_at]]] - 1); // Suggestion of Angus Johnson
    end;
  end;
end; { INext }

destructor TmTSearchO.Destroy;
begin
  inherited Destroy;
end; { Destroy }

Initialization
  MakeCompTable;

end.

